﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Numerics;
using System.Windows.Input;

namespace InterpolationTest
{
    public partial class InterpolationForm : Form
    {
        private Stopwatch sw = new Stopwatch();
        private Random rnd = new Random();
        private System.Windows.Forms.Timer timer1;

        private Vector2D cameraPos = new Vector2D(0, 0);
        private double cameraScale = 100.0;
        private bool mouseMove = false;
        private Vector2D cameraOffset = new Vector2D(0, 0);
        private Vector2D clickPos = new Vector2D(0, 0);

        private ServerSim serverSim = new ServerSim();
        private ClientSim clientSim = new ClientSim();

        public InterpolationForm()
        {
            InitializeComponent();

            base.DoubleBuffered = true;

            this.components = new System.ComponentModel.Container();
            this.timer1 = new System.Windows.Forms.Timer(this.components);
            this.timer1.Enabled = true;
            this.timer1.Interval = 1;
            this.timer1.Tick += new System.EventHandler(this.update_Tick);

            sw.Start();
        }

        private void update_Tick(object sender, System.EventArgs e)
        {
            Refresh();

            //double elapsed = (double)sw.ElapsedMilliseconds / 1000.0;
            double elapsed = (double)sw.Elapsed.TotalMilliseconds / 1000.0;
            sw.Restart();

            serverSim.update(elapsed);
            clientSim.update(elapsed);

            // simulate networking: snapshots sent from the server are transferred to the client for reception
            while (serverSim.snapshots.Count > 0)
            {
                Snapshot first = serverSim.snapshots[0];
                if (first.m_latencyTimer > 0.0f)
                {
                    // still waiting for latency delay to expire
                    // note that snapshots must be processed in order, so if the first one if still lagging behind, don't try to process the next ones
                    break;
                }

                serverSim.snapshots.RemoveAt(0);
                clientSim.addSnapshot(first.m_time, first.m_pos, first.m_vel);

                //System.Diagnostics.Debug.WriteLine("addSnaphost(first.m_time): " + first.m_time);
            }

            cameraPos = clientSim.renderPos1 + cameraOffset; // camera always centered on local player

            this.Text = "Interpolation latency = " + (int)(serverSim.currentLag * 1000.0f) + " ms | Time smoothing: " + (clientSim.smoothTimerState ? "yes" : "no");
        }

        private Vector2D worldToScreen(Vector2D pos)
        {
            double width = Width - 16;
            double height = Height - 16;

            double centerX = width * 0.5;
            double centerY = height * 0.5;

            double x = (pos.X - cameraPos.X) * cameraScale + centerX;
            double y = -(pos.Y - cameraPos.Y) * cameraScale + centerY;

            return new Vector2D(x, y);
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);

            Graphics graphics = e.Graphics;
            graphics.Clear(this.BackColor);

            // render grid
            double size = 100.0;
            for (double i = -size; i <= size + 0.001; i += 1.0)
            {
                Vector2D start = worldToScreen(new Vector2D(i, -size));
                Vector2D end = worldToScreen(new Vector2D(i, size));

                graphics.DrawLine(Pens.Blue, (float)start.X, (float)start.Y, (float)end.X, (float)end.Y);

                start = worldToScreen(new Vector2D(-size, i));
                end = worldToScreen(new Vector2D(size, i));

                graphics.DrawLine(Pens.Blue, (float)start.X, (float)start.Y, (float)end.X, (float)end.Y);
            }

            // origin
            Vector2D xy = worldToScreen(new Vector2D(0, 0));
            Rectangle rectangle = new Rectangle((int)xy.X - 8, (int)xy.Y - 8, 16, 16);
            graphics.DrawEllipse(System.Drawing.Pens.Blue, rectangle);

            // render snapshot data received
            for (int i = 0; i < clientSim.snapshots.Count - 1; ++i)
            {
                Snapshot snapshot = clientSim.snapshots[i];
                xy = worldToScreen(snapshot.m_pos);
                rectangle = new Rectangle((int)xy.X - 2, (int)xy.Y - 2, 4, 4);
                graphics.DrawRectangle(System.Drawing.Pens.Gray, rectangle);
            }

            // render snapshot data created
            for (int i = 0; i < serverSim.snapshots.Count - 1; ++i)
            {
                Snapshot snapshot = serverSim.snapshots[i];
                xy = worldToScreen(snapshot.m_pos);
                rectangle = new Rectangle((int)xy.X - 3, (int)xy.Y - 3, 6, 6);
                graphics.DrawRectangle(System.Drawing.Pens.Black, rectangle);
            }

            // render local player
            xy = worldToScreen(clientSim.localPos1);
            rectangle = new Rectangle((int)xy.X - 4, (int)xy.Y - 4, 8, 8);
            graphics.DrawEllipse(System.Drawing.Pens.Red, rectangle);

            // render local player
            xy = worldToScreen(clientSim.renderPos1);
            rectangle = new Rectangle((int)xy.X - 4, (int)xy.Y - 4, 8, 8);
            graphics.DrawEllipse(System.Drawing.Pens.Pink, rectangle);

            // render server player
            xy = worldToScreen(serverSim.serverPos2);
            rectangle = new Rectangle((int)xy.X - 4, (int)xy.Y - 4, 8, 8);
            graphics.DrawEllipse(System.Drawing.Pens.Purple, rectangle);

            // render other player
            xy = worldToScreen(clientSim.localPos2);
            rectangle = new Rectangle((int)xy.X - 4, (int)xy.Y - 4, 8, 8);
            if (clientSim.isExtrapolating)
            {
                graphics.FillEllipse(Brushes.Green, rectangle);
            }
            else
            {
                graphics.DrawEllipse(System.Drawing.Pens.Green, rectangle);
            }

            // render other player
            xy = worldToScreen(clientSim.renderPos2);
            rectangle = new Rectangle((int)xy.X - 4, (int)xy.Y - 4, 8, 8);
            if (clientSim.isExtrapolating)
            {
                graphics.FillEllipse(Brushes.Cyan, rectangle);
            }
            else
            {
                graphics.DrawEllipse(System.Drawing.Pens.Cyan, rectangle);
            }


        }

        private void InterpolationForm_MouseWheel(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            double count = (double)e.Delta * (double)SystemInformation.MouseWheelScrollLines / 120.0;
            cameraScale = cameraScale * (1.0 + count * 0.1);
        }

        private void InterpolationForm_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == 'T' || e.KeyChar == 't')
            {
                clientSim.smoothTimerState = !clientSim.smoothTimerState;
            }
            else if (e.KeyChar == 'A' || e.KeyChar == 'a')
            {
                Console.WriteLine("Resetting server velocity");
                serverSim.serverVel2 = clientSim.localVel1;
            }
            else if (e.KeyChar == '1')
            {
                clientSim.localVel1 = new Vector2D(0.0, 0.0);
                serverSim.serverVel2 = new Vector2D(0.0, 0.0);
            }
            else if (e.KeyChar == '2')
            {
                clientSim.localVel1 = new Vector2D(0.0, 0.1);
                serverSim.serverVel2 = new Vector2D(0.0, 0.1);
            }
            else if (e.KeyChar == '3')
            {
                clientSim.localVel1 = new Vector2D(0.0, 0.5);
                serverSim.serverVel2 = new Vector2D(0.0, 0.5);
            }
            else if (e.KeyChar == '4')
            {
                clientSim.localVel1 = new Vector2D(0.0, 5.0);
                serverSim.serverVel2 = new Vector2D(0.0, 5.0);
            }
        }

        private void InterpolationForm_MouseDown(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            mouseMove = true;
            clickPos = new Vector2D(e.X, e.Y);
        }

        private void InterpolationForm_MouseUp(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            mouseMove = false;
        }

        private void InterpolationForm_MouseMove(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            if (mouseMove)
            {
                cameraOffset.X -= (e.X - clickPos.X) / cameraScale;
                cameraOffset.Y += (e.Y - clickPos.Y) / cameraScale;
                clickPos = new Vector2D(e.X, e.Y);
            }
        }
    }
}
